from logging import getLogger

from selenium.webdriver import Chrome
from selenium.webdriver.common.by import By

from .humanlike import randsleep

logger = getLogger(__name__)

CAPTCHA_ATTEMPTS = 4


def solve_captcha(driver: Chrome) -> bool:
    driver.switch_to.default_content()
    iframe = driver.find_element(value="main-iframe")
    driver.switch_to.frame(iframe)
    iframe = driver.find_element(
        By.CSS_SELECTOR,
        "iframe[name*='a-'][src*='https://www.google.com/recaptcha/api2/anchor?']",
    )
    driver.switch_to.frame(iframe)
    randsleep(0.2)
    driver.find_element(By.XPATH, "//span[@id='recaptcha-anchor']").click()
    driver.switch_to.default_content()
    randsleep(0.2)
    iframe = driver.find_element(value="main-iframe")
    driver.switch_to.frame(iframe)
    if "Why am I seeing this page" in driver.page_source:
        logger.info("Completing catpcha 1")
        randsleep(0.2)

        iframe = driver.find_element(
            By.CSS_SELECTOR,
            "iframe[title*='recaptcha challenge'][src*='https://www.google.com/recaptcha/api2/bframe?']",
        )
        driver.switch_to.frame(iframe)
        randsleep(0.2)

        for _ in range(CAPTCHA_ATTEMPTS):
            logger.info("Completing catpcha")
            # let buster do it for us:
            driver.find_elements(By.CLASS_NAME, "help-button-holder")[0].click()
            randsleep(5)
            if "Multiple correct solutions required" not in driver.page_source:
                break

        driver.switch_to.default_content()
        randsleep(0.5)

    return "Why am I seeing this page" in driver.page_source



import datetime as dt
import warnings
from collections import deque
from datetime import datetime
from logging import getLogger
from time import monotonic, sleep
from typing import Callable

from pydantic import BaseModel
from selenium.webdriver import Chrome
from selenium.webdriver.common.by import By

from .browser import Browser, BrowserError
from .humanlike import randsleep, spinner_sleep
from .utils import solve_captcha


class Test(BaseModel):
    """A driving test."""

    date: datetime
    centre: str


class Centre(BaseModel):
    """A centre and optional dates.

    If dates are not supplied, the defaults from the `Driver()` are used.

    Args:
        centre (str): the centre name.
        not_before (datetime): tests must not be before this date. (Default=None)
        not_after (datetime): tests must not be after this date. (Default=None)

    """

    centre: str
    not_before: datetime = None
    not_after: datetime = None


class Driver(BaseModel):
    """A driver seeking a driving test.

    Args:
        licence_number (str): the driver's licence number.
        name (str): the driver's name.
        booking_ref (str): the driver's booking reference.
        centres (list[Centre]): the centres to search.
        not_before (datetime): do not consider tests before this date.
        not_after (datetime): do not consider tests after this date.
        disabled_dates: (list[datetime]): ignore tests on these dates.  (Default=None)
    """

    licence_number: str
    name: str
    booking_ref: str
    centres: list[Centre]
    not_before: datetime
    not_after: datetime
    disabled_dates: list[datetime] = None
    current_test: Test = None
    refresh_urls: dict = None


class ScraperError(Exception):
    pass


class BookingError(ScraperError):
    pass


class LoginError(Exception):
    pass


class Scraper:
    instances = []
    DVSA_OPENS = dt.time(6)
    DVSA_CLOSES = dt.time(23, 30)

    def __init__(
        self,
        browser: Browser,
        drivers: list[Driver],
        notify: Callable = None,
        reserve: bool = True,
        short_notice: bool = True,
        error_period: int = 30,
    ):
        self._browser = browser
        self.notify = notify or print
        i = max(self.instances) + 1 if self.instances else 1
        self.instances.append(i)
        self.name = f"Browser-{i}"
        self._logger = getLogger(self.name)
        self.drivers = drivers
        for driver in self.drivers:
            driver.refresh_urls = {}
        self._logged_in = False
        self.reserve = reserve
        self.short_notice = short_notice
        self.period = 5
        self.error_period = error_period
        self._search_counter = 0
        self.running = True
        """A Scraper which finds tests for us.

        Args:
            browser (Browser): the browser we use to do the scraping.

            drivers (list[Driver]): the drivers to find tests for.

            notify (callable): the function to call when we find something.  If
                               unset we use `print`, so you should set this.
                               Receives one argument, a str. (default=None)

            reserve (bool): whether to reserve a test if we find it. (default=True)

            short_notice (bool): whether to consider short notice tests. (default=True)

            error_period (int): time in seconds to sleep when we have an error (default=30).
        """

    def dvsa_disabled(self):
        return datetime.now() < datetime.combine(
            dt.date.today(), self.DVSA_OPENS
        ) or datetime.now() > datetime.combine(dt.date.today(), self.DVSA_CLOSES)

    @staticmethod
    def input_text_box(el, text, click=False):
        if click:
            el.click()
            randsleep(0.2)
        for char in text:
            el.send_keys(char)
            randsleep(0.01)

    @staticmethod
    def parse_timestr(timestr: str) -> datetime:
        return datetime.strptime(timestr, "%A %d %B %Y %I:%M%p")

    def logged_in(self, browser) -> bool:
        return (
            self._logged_in
            and "queue" not in browser.current_url
            and "/login" not in browser.page_source
        )

    def login(self, browser: Chrome, driver: Driver):
        if self.logged_in(browser):
            return

        browser.bypass()
        if "queue" in browser.current_url:
            self._logger.info("Queuing...")
        while "queue" in browser.current_url:
            sleep(2)
        randsleep(3)
        browser.bypass()

        el = browser.find_element(value="driving-licence-number")
        el.click()
        randsleep(1)
        self.input_text_box(el, driver.licence_number)

        el = browser.find_element(value="application-reference-number")
        el.click()
        randsleep(1)
        self.input_text_box(el, driver.booking_ref)

        browser.find_element(value="booking-login").click()
        randsleep(3)

        if "loginError=true" in browser.current_url:
            raise LoginError("Unable to login.")
        self._logged_in = True

    def find_tests(self, browser: Chrome, driver: Driver):
        self.login(browser, driver)
        main = browser.find_element(value="main").get_attribute("innerHTML")
        if "has been cancelled" in main:
            raise ScraperError("Booking Cancelled.")
        elif "has now been exceeded" in main:
            raise ScraperError("Maximum rebookings exceeded.")

        contents_container = browser.find_elements(By.CLASS_NAME, "contents")
        if not contents_container:
            raise ScraperError("Failed to find contents container.")

        current_test_date = self.parse_timestr(
            contents_container[0]
            .find_element(By.XPATH, ".//dd")
            .get_attribute("innerHTML")
        )

        current_centre = (
            contents_container[1]
            .find_element(By.XPATH, ".//dd")
            .get_attribute("innerHTML")
        )

        driver.current_test = Test(date=current_test_date, centre=current_centre)

        self._logger.debug(
            f"Current Booking for {driver.name} is on "
            f"{current_test_date} in {current_centre}"
        )
        for centre in driver.centres:
            self._logger.debug(
                f"Looking for dates for {driver.name} between "
                f"{centre.not_before or driver.not_before} and "
                f"{centre.not_after or driver.not_after}"
            )
            # wait for a bit so we don't look too automated.
            randsleep(1.5)

            # go back until we're at the starting screen.
            depth = self.find_next_available(browser, driver, centre)
            for _ in range(depth):
                randsleep(0.5)
                browser.back()

    def get_centre_url(self, browser, centre):
        browser.find_element(value="test-centre-change").click()
        randsleep(3)
        el = browser.find_element(value="test-centres-input")
        el.clear()
        self.input_text_box(el, centre)
        browser.find_element(value="test-centres-submit").click()
        randsleep(5)
        container = browser.find_element(By.CLASS_NAME, "test-centre-results")
        centre = container.find_element(By.XPATH, ".//a")
        return centre.get_attribute("href")

    def find_next_available(self, browser: Chrome, driver: Driver, centre: Centre):
        # try to avoid hitting search limit for day
        randsleep(18)
        self._search_counter += 1
        self._logger.info(
            f"Search {self._search_counter}: Trying centre {centre.centre}"
        )
        if centre.centre == driver.current_test.centre:
            back = 2
            browser.find_element(value="date-time-change").click()
            randsleep(1)
            browser.find_element(value="test-choice-earliest").click()
            randsleep(1)
            browser.execute_script("window.scrollTo(0, document.body.scrollHeight);")
            randsleep(1)
            browser.find_element(value="driving-licence-submit").click()
            randsleep(1)
        else:
            back = 3
            refresh_url = driver.refresh_urls.get(
                centre.centre, self.get_centre_url(browser, centre.centre)
            )
            browser.get(refresh_url)
            randsleep(1)

        page = browser.page_source
        if any(
            x in page
            for x in (
                "Oops",
                "in the queue",
                "Incapsula incident",
                "Enter details below",
                "Search limit reached",
            )
        ):
            browser.bypass()  # is this needed now we handle on getting?

        if "no tests available" in page:
            self._logger.info("No tests available.")
            return back

        day, el = self._scan_for_test(browser, driver, centre)
        if not day:
            self._logger.info("No tests in required range.")
            return back

        if self.reserve:
            slot = self._reserve_test(browser, day, centre, el)
            if slot:
                self._logger.info(f"Reserved test: {slot}")
            else:
                self._logger.info("Failed to reserve test...")
        else:
            self.notify(f"Test found at {centre} on {day}")
            input("Deal with test; press enter when done.")
        return back

    @staticmethod
    def correct_month_showing(browser: Chrome, day: datetime):
        el = browser.find_element(By.CLASS_NAME, "BookingCalendar-currentMonth")
        return day.strftime("%B") in el.get_attribute("innerHTML")

    def _reserve_test(self, browser: Chrome, day: datetime, centre: str, el) -> bool:
        """Reserve a test.  **UNTESTED**"""
        warnings.warn(
            "Attempting to reserve test: code untested.  YMMV!", RuntimeWarning
        )

        # scroll to correct month
        attempts = 0
        while not self.correct_month_showing(browser, day):
            browser.find_element(By.CLASS_NAME, "BookingCalendar-nav--prev").click()
            attempts += 1
            if attempts > 12:
                raise BookingError("Failed to find correct month.")

        # click on date.
        el.click()
        # get container
        time_container = browser.find_element(By.ID, f"date-{day.strftime('%Y-%m-%d')}")
        label = time_container.find_element(By.XPATH, ".//label")
        # get time
        time_ms = int(label.get_attribute("for").replace("slot-", ""))
        time = datetime.fromtimestamp(time_ms / 1000).time()
        test_slot = datetime.combine(day.date(), time)
        self._logger.info(f"earliest availabe slot is ", test_slot)

        # check if short_notice
        try:
            short_notice = (
                label.get_attribute("for").get_attribute("data-short-notice") == "true"
            )
        except Exception:
            short_notice = False

        # add error handling here if required.
        randsleep(1)
        # click on time box
        first_available_time = time_container.find_element(By.XPATH, "./label")
        randsleep(1)
        first_available_time.click()
        randsleep(1)
        # click on continue button
        continue_button = browser.find_element(By.ID, "slot-chosen-submit")
        randsleep(1)
        continue_button.click()
        # label.click()
        randsleep(1)

        # dismiss warning
        if short_notice:
            if self.short_notice:
                browser.find_element(
                    By.XPATH, "(//button[@id='slot-warning-continue'])[2]"
                ).click()
            else:
                self._logger.info("Skipping test as short notice.")
                return None
        else:
            # click submit button
            slot_warning_continue_button = browser.find_element(
                By.ID, "slot-warning-continue"
            )
            slot_warning_continue_button.click()
        randsleep(5)

        self._logger.info("15 min timer started, releases in about 16.5 min ")
        # this is wrapped in a loop in the original code.  I'm not sure what those multiple attempts are for.

        randsleep(3)
        # we are the candidate
        browser.find_element(By.ID, "i-am-candidate").click()
        randsleep(1)

        # we make no manual attempt to solve the captcha here.  It might be
        # solved for us by the Browser().
        if "no longer available" in browser.page_source:
            self._logger.info("Missed it: someone else got there first...")
            return False
        self.notify(f"A Driving test has been reserved on {test_slot} in {centre}.")

        # start of new hold loop
        # sleep for long time while dvsa release back to pool
        spinner_sleep(950)
        randsleep(15)

        # find and click abandon button
        abandon_button = browser.find_element(By.ID, "abandon")
        randsleep(1)
        abandon_button.click()
        # TODO add explicit wait until clickable, something like
        # element = WebDriverWait(driver, 10).until \
        #     (EC.element_to_be_clickable((By.ID, "test-centre-change")))
        # driver.find_element_by_id("test-centre-change").click()

        randsleep(2)
        # find and click abandon change button
        abandon_changes_button = browser.find_element(By.ID, "abandon-changes")
        randsleep(1)
        abandon_changes_button.click()
        randsleep(1)
        self._logger.info("Released test.")
        randsleep(2)
        return test_slot

    def _scan_for_test(self, browser: Chrome, driver: Driver, centre: Centre):
        cal = browser.find_element(By.CLASS_NAME, "BookingCalendar-datesBody")
        days = cal.find_elements(By.XPATH, ".//td")
        for day in days:
            if "--unavailable" in day.get_attribute("class"):
                continue
            a = day.find_element(By.XPATH, ".//a")
            date = datetime.strptime(a.get_attribute("data-date"), "%Y-%m-%d")

            if driver.disabled_dates and date in driver.disabled_dates:
                continue

            not_before = centre.not_before or driver.not_before
            if date < not_before:
                continue

            # prevent it finding a test before the current test if the current
            # test is in the required window.
            if driver.current_test.date > not_before:

                # We probably don't care to move back test by test on the same day,
                # so we only accept tests on at least the day before.
                if date.date() >= driver.current_test.date.date():
                    continue

            not_after = centre.not_after or driver.not_after
            if date > not_after:
                continue
            return date, a

        return None, None

    def time_to_next_event(self, elapsed: dt.timedelta = None) -> int:
        """Time to sleep between runs.

        Returns:
            time (int): time in seconds.
        """
        subtract = elapsed.seconds if elapsed else 0
        return max(self.period * 60 - subtract, 0)

    def error_callback(self):
        """Run something when we have an error."""
        pass

    def loop_sleep(self, period):
        """Sleep fn for main loop."""
        period = max(period - 5, 0)
        self._logger.debug(f"Sleeping for ~{period}s.")
        if period:
            randsleep(period, period + 45)
        self._logger.debug(f"Sleeping for ~5s to introduce some randomness.")
        randsleep(5)

    def __call__(self):
        while True:
            self._logger.info("Starting search loop")

            if self.dvsa_disabled():
                self._logger.info("Website not currently available; sleeping.")

            while self.dvsa_disabled():
                randsleep(60)  # has spinner

            while not self.running:
                randsleep(60)

            printed = False
            errs = deque([], maxlen=5)
            searched, errored = 0, 0

            try:
                with self._browser as browser:
                    while self.running:
                        for driver in self.drivers:
                            start = datetime.now()
                            self._logger.info(
                                f"Finding tests for driver {driver.name} "
                                f"run {searched +1}, of which {errored} errored."
                            )
                            self.find_tests(browser, driver)
                            end = datetime.now()
                            period = self.time_to_next_event(elapsed=end - start)
                            self.loop_sleep(period)
                            searched += 1
            except Exception as e:
                errored += 1
                errs.append(monotonic())
                self._logger.exception(e)
                self.error_callback()
                randsleep(self.error_period)

            if len(errs) == 5 and errs[-1] - errs[0] < 10 * 60:
                msg = (
                    "Five errors within 10 minutes; "
                    "giving up so this IP doesn't get blacklisted."
                )
                self._logger.error(msg)  # make sure it's displayed, however we handle
                raise ScraperError(msg)



"""Browser and CaptchaChrome objects based around `undetected_chromedriver`."""
import multiprocessing
from logging import getLogger
from pathlib import Path

import undetected_chromedriver as uc
from selenium.webdriver.common.by import By

from .browser import Browser, BrowserError, CaptchaChromeBase, TorBrowser

logger = getLogger(__name__)

# Monkeypatch undetected_chromedriver
# I'd love to know why we have to use shell=True to connect properly...


def _start_detached(executable, *args, writer: multiprocessing.Pipe = None):
    kwargs = {}
    if uc.dprocess.platform.system() == "Windows":
        kwargs.update(
            creationflags=uc.dprocess.DETACHED_PROCESS
            | uc.dprocess.CREATE_NEW_PROCESS_GROUP
        )
    elif uc.dprocess.sys.version_info < (3, 2):
        # assume posix
        kwargs.update(preexec_fn=uc.dprocess.os.setsid)
    else:  # Python 3.2+ and Unix
        kwargs.update(start_new_session=True)

    # run
    if not executable.startswith('"'):
        executable = f'"{executable}"'
    cmdline = " ".join([executable, *args])
    logger.debug(f"running chrome, cmdline is {cmdline}")
    PIPE = uc.dprocess.PIPE
    p = uc.dprocess.Popen(
        cmdline, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True, **kwargs
    )
    writer.send(p.pid)
    uc.sys.exit()


uc.dprocess._start_detached = _start_detached


class UndetectedCaptchaChrome(CaptchaChromeBase, uc.Chrome):
    """A CaptchaChrome based on `undetected_chromedriver`."""


class UndetectedBrowser(Browser):
    """A browser which lets `undetected_chromedriver` launch chrome for us."""

    def __init__(self, *args, chrome=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._chrome = f'"{chrome}"' if chrome else None

    def launch_chrome(self):
        pass

    def _connect(self, *args, chrome_options: uc.ChromeOptions = None):
        chrome_options = chrome_options or uc.ChromeOptions()
        if self._buster:
            chrome_options.add_argument(self.buster_arg)

        kwargs = {"options": chrome_options, "browser_executable_path": self._chrome}
        driver = UndetectedCaptchaChrome(**kwargs)
        self._proc = uc.dprocess.REGISTERED[-1]

        driver.get(self._url)
        return driver


class UndetectedTorBrowser(UndetectedBrowser, TorBrowser):
    # def start_tor(self):
    #     self._tor_port = 8897
    def launch_chrome(self):
        self.start_tor()

    @property
    def tor_arg(self):
        return f'--proxy-server="socks4://localhost:{self.tor_port}"'

    def _connect(self, *args, chrome_options: uc.ChromeOptions = None):
        chrome_options = chrome_options or uc.ChromeOptions()
        chrome_options.add_argument(self.tor_arg)
        return super()._connect(*args, chrome_options=chrome_options)


class ManualBusterMixin:
    BUSTER_URL = "https://chrome.google.com/webstore/detail/buster-captcha-solver-for/mpbjkejclgfgadiemmefgebjfooflfhl?hl=en"
    PROFILE_DIR = "blank-profile"

    def __init__(self, *args, profile_dir: Path = None, **kwargs):
        buster = kwargs.get("buster", None)
        if buster:
            del kwargs["buster"]
        super().__init__(*args, **kwargs)
        if buster:
            self._buster = buster
        self.profile_dir = profile_dir or self.PROFILE_DIR

    def generate_profile(self):
        """Generate a profile and then install buster manually."""
        chrome_options = uc.ChromeOptions()
        chrome_options.user_data_dir = self.profile_dir
        driver = UndetectedCaptchaChrome(options=chrome_options)
        driver.get(self.BUSTER_URL)
        driver.find_element(By.XPATH, "//*[text()='I agree']").click()
        driver.find_element(
            By.CSS_SELECTOR, ".g-c-R.webstore-test-button-label"
        ).click()
        input("Accept installation and press enter: ")
        driver.close()
        del driver

    def _connect(self, *args, chrome_options: uc.ChromeOptions = None):
        chrome_options = chrome_options or uc.ChromeOptions()
        if self._buster:
            profile_dir = Path(self.profile_dir)
            if not profile_dir.is_dir():
                if profile_dir.exists():
                    raise BrowserError(
                        f"Profile dir {profile_dir} exists and is not a directory!"
                    )
                self.generate_profile()
            chrome_options.user_data_dir = self.PROFILE_DIR
            self._buster = None
        return super()._connect(*args, chrome_options=chrome_options)


class ManualBusterUndetectedBrowser(ManualBusterMixin, UndetectedBrowser):
    """A browser which installs buster manually and then uses the same profile dir."""


class ManualBusterUndetectedTorBrowser(ManualBusterMixin, UndetectedTorBrowser):
    """A tor browser which installs buster manually and then uses the same profile dir."""



import signal
import socketserver
import traceback as tb
from datetime import datetime
from logging import getLogger
from pathlib import Path
from subprocess import PIPE, Popen
from tempfile import TemporaryDirectory
from typing import Union
from time import sleep

from fake_useragent import UserAgent
from psutil import Process
from selenium import webdriver
from selenium.webdriver.common.by import By

from .humanlike import randsleep


class BrowserError(Exception):
    pass


class ControlBrowserException(Exception):
    pass


class CaptchaChromeBase:
    """A Chrome which will try to solve captchas for you."""

    INCAPSULA = "unsuccessful. Incapsula"
    BLOCKED = "Access denied"
    SUSPICIOUS = "Pardon our interruption"
    CAPTCHA_ATTEMPTS = 4
    WAIT_MSG = "Please wait while we process your request"
    MAX_WAIT = 60
    LONG_SLEEP = 60
    instances = []

    def __init__(self, *args, **kwargs):
        i = max(self.instances) + 1 if self.instances else 1
        self.instances.append(i)
        name = f"CaptchaChrome-{i}"
        self._logger = getLogger(name)
        super().__init__(*args, **kwargs)

    def solve_captcha(self) -> bool:
        if self.INCAPSULA not in self.page_source:
            return True

        self.switch_to.default_content()
        iframe = self.find_element(value="main-iframe", bypass=False)
        self.switch_to.frame(iframe)
        iframe = self.find_element(
            By.CSS_SELECTOR,
            "iframe[name*='a-'][src*='https://www.google.com/recaptcha/api2/anchor?']",
            bypass=False,
        )
        self.switch_to.frame(iframe)
        randsleep(0.2)
        self.find_element(
            By.XPATH, "//span[@id='recaptcha-anchor']", bypass=False
        ).click()
        self.switch_to.default_content()
        randsleep(0.2)
        iframe = self.find_element(value="main-iframe", bypass=False)
        self.switch_to.frame(iframe)
        if "Why am I seeing this page" in self.page_source:
            self._logger.info("Completing catpcha 1")
            randsleep(0.2)

            iframe = self.find_element(
                By.CSS_SELECTOR,
                "iframe[title*='recaptcha challenge'][src*='https://www.google.com/recaptcha/api2/bframe?']",
                bypass=False,
            )
            self.switch_to.frame(iframe)
            randsleep(0.2)

            for _ in range(self.CAPTCHA_ATTEMPTS):
                self._logger.info("Completing catpcha")
                # let buster do it for us:
                self.find_elements(By.CLASS_NAME, "help-button-holder")[0].click()
                randsleep(5)
                if "Multiple correct solutions required" not in self.page_source:
                    break

            self.switch_to.default_content()
            randsleep(0.5)

        return "Why am I seeing this page" in self.page_source

    def bypass(self):
        """Try to bypass security."""

        solved = False
        if any(x in self.page_source for x in (self.BLOCKED, self.SUSPICIOUS)):
            self._logger.error("Blocked")
            raise BrowserError("Blocked")

        elif self.INCAPSULA in self.page_source:
            self._logger.info("Trying to solve captcha.")

            for _ in range(2):

                if (
                    self.INCAPSULA in self.page_source
                    and "iframe" not in self.page_source
                ):
                    randsleep(5)
                    self.refresh()
                else:
                    solved = True

                if self.INCAPSULA in self.page_source:
                    randsleep(2)
                    self.solve_captcha()
                else:
                    solved = True

                if self.INCAPSULA in self.page_source:
                    self._logger.debug("Long sleep as incapsula still in page source.")
                    randsleep(self.LONG_SLEEP)
                else:
                    solved = True
        else:
            return

        if solved:
            self._logger.info("Solved Captcha!")
        else:
            self._logger.error("Unable to beat Captcha")
            raise BrowserError("Unable to beat Captcha")

    def wait(self):
        waited = 0
        if self.WAIT_MSG in self.page_source:
            self._logger.debug("Waiting as page is not ready yet.")
            while self.WAIT_MSG in self.page_source and waited < self.MAX_WAIT:
                sleep(1)
                waited += 1
            if self.WAIT_MSG in self.page_source:
                raise BrowserError(f"Page failed to load within {self.MAX_WAIT} s")
        else:
            self._logger.debug("No need to wait for page.")

    def find_element(self, by=None, value=None, bypass=True, wait=True):
        if bypass:
            self.bypass()
        if wait:
            self.wait()

        args = dict(by=by, value=value)
        return super().find_element(**{k: v for k, v in args.items() if v})


class CaptchaChrome(CaptchaChromeBase, webdriver.Chrome):
    """Captcha Chrome based on webdriver.Chrome."""


class Browser:
    """A browser driven by a driver over the debug interface.

    We start the browser *first* and navigate to the login page whilst the
    browser is 'clean', i.e. not controlled remotely.  This allows DVSA's
    profiling JS to run against a pretty clean browser.  We then block the
    profiling script, which seems to work, at least for a short while.
    (Otherwise we just get blocked.)"""

    URL = "https://driverpracticaltest.dvsa.gov.uk/login"
    instances = []

    def __init__(
        self,
        port: int = None,
        buster: Union[Path, str] = None,
        chrome: Union[Path, str] = "google-chrome-stable",
        chromedriver: Union[Path, str] = "chromedriver",
        url: str = None,
        errors_dir: Union[Path, str] = None,
        dump_on_error: bool = True,
        register_sigusr1: bool = True,
    ):
        """Setup the Browser.

        Args:

            port (int): the port to use for remote management.  (Default: 8745)

            buster (Path | str): path to the unzipped buster extension.  If you don't
                           provide this, captchas will be unsolveable.

            chrome (Path | str): path to the Chrome executable on your system.

            chromedrive_path(Path | str): path to the chromedriver executable on your system.

            url (str): the url to point the browser to, if not the default DVSA one.

            errors_dir (Path | str): where to save errors.

            dump_on_error (bool): whether to save error dumps for debugging.
                                  (Default: True)

            register_sigusr1: bool = True: whether to register a handler for
                                           sigusr1 which calls
                                           `self.control_browser()` (Default:
                                           True).
        """
        self._installed = False
        self._port = port
        if buster:
            buster = Path(buster)
            self.validate_buster(buster)
        self._buster = buster
        self._profile_dir = None
        self._chrome = f'"{chrome}"'
        i = max(self.instances) + 1 if self.instances else 1
        self.instances.append(i)
        self.name = f"Browser-{i}"
        self._logger = getLogger(self.name)
        self._proc = None
        self._chromedriver = chromedriver
        self._url = url or self.URL
        if errors_dir:
            errors_dir = Path(errors_dir)
            self._errors_dir = errors_dir
        else:
            self._errors_dir = Path(f"./errors/{self.name}")
        self.dump_on_error = dump_on_error
        if register_sigusr1 and hasattr(signal, "SIGUSR1"):
            signal.signal(signal.SIGUSR1, self._signal_handler)
            self._logger.debug(
                "Registered signal handler; send sigusr1 to get control."
            )

    @staticmethod
    def validate_buster(buster: Path):
        if not buster.is_dir():
            raise BrowserError("Please unzip buster and pass the dir.")

    @property
    def port(self) -> int:
        if not self._port:
            with socketserver.TCPServer(("localhost", 0), None) as s:
                self._port = s.server_address[1]
        return self._port

    @property
    def buster_arg(self) -> str:
        if self._buster:
            return f'--load-extension="{self._buster}"'
        else:
            return ""

    @property
    def profile_arg(self) -> str:
        if not self._profile_dir:
            self._profile_dir = TemporaryDirectory()
        return f"--user-data-dir={self._profile_dir.name}"

    @property
    def port_arg(self) -> str:
        return f"--remote-debugging-port={self.port}"

    def control_browser(self):
        """Take control of the browser for manual use.

        This cannot be undone and the browser will die when finished."""
        raise ControlBrowserException("User requested manual control.")

    def _signal_handler(self, sig, frame):
        self.control_browser()

    def kill(self, proc, name):
        if not proc:
            return
        self._logger.info(f"Killing {name}")

        if isinstance(proc, int):
            sysproc = Process(proc)
        else:
            sysproc = Process(proc.pid)

        for child in sysproc.children(recursive=True):
            try:
                child.terminate()
            except Exception:
                pass

        for child in sysproc.children(recursive=True):
            try:
                child.kill()
            except Exception:
                pass

        if isinstance(proc, int):
            return

        proc.terminate()
        proc.wait(2)
        if proc.poll() is None:
            self._logger.info("Failed to die: killing with SIGKILL")
            proc.kill()
            proc.wait(2)
            if proc.poll() is None:
                raise BrowserError("Process failed to die")

    def launch_chrome(self, extra_args=None):
        self._logger.info("Starting chrome")
        ua = UserAgent().random
        self._logger.info(f"This time we are {ua}")
        cmd = [
            self._chrome,
            self.port_arg,
            self.profile_arg,
            self.buster_arg,
            "--no-first-run",
            "--blink-settings=imagesEnabled=false",
            f'--user-agent="{ua}"',
            self._url,
        ]
        if extra_args:
            cmd = cmd[:1] + extra_args + cmd[1:]
        # needed at present to prevent failure
        # chrome_options.experimental_options.pop("excludeSwitches")

        # Needed to launch in the right env with a new config
        # TODO: figure out what's actually happening here and fix this.
        self._proc = Popen(" ".join(cmd), shell=True)
        self._logger.info("Waiting 10s for page to load and js to run.")
        randsleep(10)

    def _connect(self):
        """Connect to running browser."""
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_experimental_option(
            "debuggerAddress", f"localhost:{self._port}"
        )

        driver = CaptchaChrome(
            executable_path=self._chromedriver,
            options=chrome_options,
        )

        driver.execute_cdp_cmd(
            "Network.setBlockedURLs",
            {
                "urls": [
                    "https://driverpracticaltest.dvsa.gov.uk/nions-to-vnse-the-Bewarfish-so-like-here-hoa-Mon"
                ]
            },
        )
        driver.execute_cdp_cmd("Network.enable", {})
        driver.refresh()
        return driver

    def __enter__(self) -> CaptchaChrome:
        self.launch_chrome()
        self._driver = self._connect()
        return self._driver

    @staticmethod
    def _errorfn(now: datetime = None, suffix: str = "txt"):
        now = now or datetime.now()
        return Path(f"error-{now.strftime('%Y-%m-%d_%H-%M-%S')}.{suffix}")

    def _dump(self, *err):
        self._errors_dir.mkdir(parents=True, exist_ok=True)
        outf = self._errors_dir / self._errorfn()
        with outf.open("w") as f:
            f.write("".join(tb.format_exception(*err)))
            f.write("\n")
            try:
                f.write("Whilst visiting:")
                f.write(self._driver.current_url)
            except AttributeError:
                pass

        try:
            with outf.with_suffix(".html").open("w") as f:
                f.write(self._driver.page_source)
            self._driver.save_screenshot(str(outf.with_suffix(".png")))
        except AttributeError:
            pass

        self._logger.info(f"Saved error dump in {outf}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ControlBrowserException:
            input("Press enter to exit.")
        if exc_val and self.dump_on_error:
            try:
                self._dump(exc_type, exc_val, exc_tb)
            except Exception as e:
                tb.print_exc()
        self.kill(self._proc, "Chrome")
        try:
            self._driver.quit()
        except Exception:
            pass

        try:
            self._profile_dir.cleanup()
        except Exception as e:
            self._logger.exception(e)
        return False if exc_val else None


class TorBrowser(Browser):
    """A browser tunneled by TOR.

    We start the `tor` daemon ourselves so as to have a random IP. This will
    *only* work if you can run `tor` as your user. This will require removing
    the `User` option in the default `torrc`.

    The exit handler here kills the tor process, so we get a new ip for next
    time. This seems to reduce (although not eliminate?) getting profile
    blocked.
    """

    def __init__(self, tor: Union[Path, str] = "tor", tor_port: int = None, **kwargs):
        """Setup the TorBrowser.

        Args:
            tor (Path | str): The `tor` executable to call.  (Default: tor)

        See also the documentation for `Browser.__init__()`.
        """
        super().__init__(**kwargs)
        self._tor = Path(tor)
        self._tor_port = tor_port
        self._tor_proc = None

    @property
    def tor_port(self) -> str:
        if not self._tor_port:
            with socketserver.TCPServer(("localhost", 0), None) as s:
                self._tor_port = str(s.server_address[1])
        return self._tor_port

    def start_tor(self):
        self._logger.info("Starting Tor")
        cmd = [self._tor, "--SocksPort", self.tor_port]
        proc = Popen(cmd, stdout=PIPE, encoding="utf8")
        self._tor_proc = proc
        self._logger.info("Started Tor")

    def launch_chrome(self):
        self.start_tor()
        tor_args = [
            f'--proxy-server="socks4://localhost:{self.tor_port}"',
            # disable prefetch, as for some reason it's not proxied (?!)
            "--dns-prefetch-disable"
            # apparently this should be more resilient than merely disabling prefetch
            # but it doesn't seem to work
            # '--host-resolver-rules="MAP * ~NOTFOUND , EXCLUDE myproxy"',
        ]
        super().launch_chrome(extra_args=tor_args)

    def __exit__(self, *args):
        ret = super().__exit__(*args)
        self.kill(self._tor_proc, "Tor")
        return ret
